from django.shortcuts import render
from django.urls import path, reverse
from django.views.decorators.csrf import csrf_exempt
from graphene_django.views import GraphQLView

from .settings import grapple_settings


def graphiql(request):
    graphiql_settings = {
        "REACT_VERSION": "18.2.0",
        "GRAPHIQL_VERSION": "2.4.7",
        "endpointURL": reverse("grapple_graphql"),
        "supports_subscriptions": False,
    }
    return render(request, "grapple/graphiql.html", graphiql_settings)


# Traditional URL routing
urlpatterns = [
    path("graphql/", csrf_exempt(GraphQLView.as_view()), name="grapple_graphql")
]

if grapple_settings.EXPOSE_GRAPHIQL:
    urlpatterns.append(path("graphiql/", graphiql, name="grapple_graphiql"))



import graphene
from graphql import GraphQLError
from wagtail.contrib.settings.models import BaseGenericSetting, BaseSiteSetting
from wagtail.models import Site

from ..registry import registry
from ..utils import resolve_site_by_hostname


def SettingsQuery():
    if registry.settings:

        class SettingsObjectType(graphene.Union):
            class Meta:
                types = registry.settings.types

        class Mixin:
            setting = graphene.Field(
                SettingsObjectType,
                name=graphene.String(),
                site=graphene.String(),
            )
            settings = graphene.List(
                graphene.NonNull(SettingsObjectType),
                required=True,
                name=graphene.String(),
                site=graphene.String(),
            )

            # Return just one setting base on name param.
            def resolve_setting(self, info, **kwargs):
                # Site filter
                # Only applies to settings that inherit from BaseSiteSetting
                site_hostname_kwarg = "site"
                site_hostname = kwargs.pop(site_hostname_kwarg, None)

                if site_hostname is not None:
                    site = resolve_site_by_hostname(
                        hostname=site_hostname,
                        filter_name=site_hostname_kwarg,
                    )
                else:
                    site = None

                name = kwargs.get("name")
                for setting in registry.settings:
                    # If 'name' filter used, ignore any models that don't match the filter
                    if name and setting._meta.model_name != name.lower():
                        continue

                    if issubclass(setting._meta.model, BaseSiteSetting):
                        if site:
                            return setting._meta.model.objects.filter(site=site).first()
                        elif Site.objects.all().count() == 1:
                            # If there's only one Site, we can reliably return
                            # the correct (i.e. only) SiteSetting.
                            return setting._meta.model.objects.first()
                        else:
                            # If there are multiple `Site`s, we don't know what
                            # data to return.
                            raise GraphQLError(
                                f"There are multiple `{name}` instances - "
                                "please include a `site` filter to disambiguate "
                                f"(e.g. `setting(name: '{name}', site='example.com')`."
                            )

                    elif issubclass(setting._meta.model, BaseGenericSetting):
                        # If it's a GenericSetting, there can only be one.
                        return setting._meta.model.objects.first()

                    return None

            # Return all settings.
            def resolve_settings(self, info, **kwargs):
                # Site filter
                # Only applies to settings that inherit from BaseSiteSetting
                site_hostname_kwarg = "site"
                site_hostname = kwargs.pop(site_hostname_kwarg, None)

                if site_hostname is not None:
                    site = resolve_site_by_hostname(
                        hostname=site_hostname,
                        filter_name=site_hostname_kwarg,
                    )
                else:
                    site = None

                name = kwargs.get("name")
                settings_objects = []
                for setting in registry.settings:
                    # If 'name' filter used, ignore any models that don't match the filter
                    if name and setting._meta.model_name != name.lower():
                        continue

                    if site and issubclass(setting._meta.model, BaseSiteSetting):
                        settings_objects.extend(
                            setting._meta.model.objects.filter(site=site)
                        )

                    else:
                        settings_objects.extend(setting._meta.model.objects.all())

                return settings_objects

        return Mixin

    else:

        class Mixin:
            pass

        return Mixin



class IllegalDeprecation(Exception):
    """
    A Collection or Field was marked as required, but also received a `deprecation_reason'.

    This is invalid - a deprecated entity must be nullable.
    """

    pass



import inspect

import graphene

# TODO: use specific imports
import wagtail
import wagtail.documents.blocks
import wagtail.embeds.blocks
import wagtail.images.blocks
import wagtail.snippets.blocks
from graphene.types import Scalar
from graphene_django.converter import convert_django_field
from wagtail import blocks
from wagtail.embeds.blocks import EmbedValue
from wagtail.embeds.embeds import get_embed
from wagtail.embeds.exceptions import EmbedException
from wagtail.fields import StreamField
from wagtail.rich_text import RichText

from ..registry import registry
from .rich_text import RichText as RichTextType


class GenericStreamFieldInterface(Scalar):
    @staticmethod
    def serialize(stream_value):
        try:
            return stream_value.raw_data
        except AttributeError:
            return stream_value.stream_data


@convert_django_field.register(StreamField)
def convert_stream_field(field, registry=None):
    return GenericStreamFieldInterface(
        description=field.help_text, required=not field.null
    )


class StreamFieldInterface(graphene.Interface):
    id = graphene.String()
    block_type = graphene.String(required=True)
    field = graphene.String(required=True)
    raw_value = graphene.String(required=True)

    @classmethod
    def resolve_type(cls, instance, info):
        """
        If block has a custom Graphene Node type in registry then use it,
        otherwise use generic block type.
        """
        if hasattr(instance, "block"):
            mdl = type(instance.block)
            if mdl in registry.streamfield_blocks:
                return registry.streamfield_blocks[mdl]

            for block_class in inspect.getmro(mdl):
                if block_class in registry.streamfield_blocks:
                    return registry.streamfield_blocks[block_class]

        return registry.streamfield_blocks["generic-block"]

    def resolve_id(self, info, **kwargs):
        return self.id

    def resolve_block_type(self, info, **kwargs):
        return type(self.block).__name__

    def resolve_field(self, info, **kwargs):
        return self.block.name

    def resolve_raw_value(self, info, **kwargs):
        if isinstance(self, blocks.StructValue):
            # This is the value for a nested StructBlock defined via GraphQLStreamfield
            return serialize_struct_obj(self)
        elif isinstance(self.value, dict):
            return serialize_struct_obj(self.value)
        elif isinstance(self.value, RichText):
            # Ensure RichTextBlock raw value always returns the "internal format", rather than the conterted value
            # as per https://docs.wagtail.io/en/stable/extending/rich_text_internals.html#data-format.
            # Note that RichTextBlock.value will be rendered HTML by default.
            return self.value.source

        return self.value


def generate_streamfield_union(graphql_types):
    class StreamfieldUnion(graphene.Union):
        class Meta:
            types = graphql_types

        @classmethod
        def resolve_type(cls, instance, info):
            """
            If block has a custom Graphene Node type in registry then use it,
            otherwise use generic block type.
            """
            mdl = type(instance.block)
            if mdl in registry.streamfield_blocks:
                return registry.streamfield_blocks[mdl]

            return registry.streamfield_blocks["generic-block"]

    return StreamfieldUnion


class StructBlockItem:
    id = None
    block = None
    value = None

    def __init__(self, id, block, value=""):
        self.id = id
        self.block = block
        self.value = value


def serialize_struct_obj(obj):
    rtn_obj = {}

    if hasattr(obj, "raw_data"):
        rtn_obj = []
        for field in obj[0]:
            rtn_obj.append(serialize_struct_obj(field.value))
    # This conditionnal and below support both Wagtail >= 2.13 and <2.12 versions.
    # The "stream_data" check can be dropped once 2.11 is not supported anymore.
    # Cf: https://docs.wagtail.io/en/stable/releases/2.12.html#stream-data-on-streamfield-values-is-deprecated
    elif hasattr(obj, "stream_data"):
        rtn_obj = []
        for field in obj.stream_data:
            rtn_obj.append(serialize_struct_obj(field["value"]))
    else:
        for field in obj:
            value = obj[field]
            if hasattr(value, "raw_data"):
                rtn_obj[field] = [serialize_struct_obj(data.value) for data in value[0]]
            elif hasattr(obj, "stream_data"):
                rtn_obj[field] = [
                    serialize_struct_obj(data["value"]) for data in value.stream_data
                ]
            elif hasattr(value, "value"):
                rtn_obj[field] = value.value
            elif hasattr(value, "src"):
                rtn_obj[field] = value.src
            elif hasattr(value, "file"):
                rtn_obj[field] = value.file.url
            else:
                rtn_obj[field] = value

    return rtn_obj


class StructBlock(graphene.ObjectType):
    class Meta:
        interfaces = (StreamFieldInterface,)

    blocks = graphene.List(graphene.NonNull(StreamFieldInterface), required=True)

    def resolve_blocks(self, info, **kwargs):
        stream_blocks = []

        if issubclass(type(self.value), blocks.stream_block.StreamValue):
            # self: StreamChild, block: StreamBlock, value: StreamValue
            stream_data = self.value[0]
            child_blocks = self.value.stream_block.child_blocks
        else:
            # This occurs when StreamBlock is child of StructBlock
            # self: StructBlockItem, block: StreamBlock, value: list
            stream_data = self.value
            child_blocks = self.block.child_blocks

        for field, value in stream_data.items():
            block = dict(child_blocks)[field]
            if issubclass(type(block), blocks.ChooserBlock) or not issubclass(
                type(block), blocks.StructBlock
            ):
                if isinstance(value, int):
                    value = block.to_python(value)

            stream_blocks.append(StructBlockItem(field, block, value))

        return stream_blocks


class StreamBlock(StructBlock):
    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_blocks(self, info, **kwargs):
        child_blocks = self.value.stream_block.child_blocks

        return [
            StructBlockItem(
                id=stream.id, block=child_blocks[stream.block_type], value=stream.value
            )
            for stream in self.value
        ]


class StreamFieldBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class CharBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class TextBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class EmailBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class IntegerBlock(graphene.ObjectType):
    value = graphene.Int(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class FloatBlock(graphene.ObjectType):
    value = graphene.Float(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class DecimalBlock(graphene.ObjectType):
    value = graphene.Float(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class RegexBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class URLBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class BooleanBlock(graphene.ObjectType):
    value = graphene.Boolean(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class DateBlock(graphene.ObjectType):
    value = graphene.String(format=graphene.String(), required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_value(self, info, **kwargs):
        format = kwargs.get("format")
        if format:
            return self.value.strftime(format)
        return self.value


class DateTimeBlock(DateBlock):
    class Meta:
        interfaces = (StreamFieldInterface,)


class TimeBlock(DateBlock):
    class Meta:
        interfaces = (StreamFieldInterface,)


class RichTextBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_value(self, info, **kwargs):
        return RichTextType.serialize(self.value.source)


class RawHTMLBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class BlockQuoteBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class ChoiceOption(graphene.ObjectType):
    key = graphene.String(required=True)
    value = graphene.String(required=True)


class ChoiceBlock(graphene.ObjectType):
    value = graphene.String(required=True)
    choices = graphene.List(graphene.NonNull(ChoiceOption), required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_choices(self, info, **kwargs):
        choices = []
        for key, value in self.block._constructor_kwargs["choices"]:
            choice = ChoiceOption(key, value)
            choices.append(choice)
        return choices


def get_embed_url(instance):
    return instance.value.url if hasattr(instance, "value") else instance.url


def get_embed_object(instance):
    try:
        return get_embed(get_embed_url(instance))
    except EmbedException:
        pass


class EmbedBlock(graphene.ObjectType):
    value = graphene.String(required=True)
    url = graphene.String(required=True)
    embed = graphene.String()
    raw_embed = graphene.JSONString()

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_url(self, info, **kwargs):
        return get_embed_url(self)

    def resolve_raw_value(self, info, **kwargs):
        if isinstance(self, EmbedValue):
            return self
        return StreamFieldInterface.resolve_raw_value(info, **kwargs)

    def resolve_embed(self, info, **kwargs):
        embed = get_embed_object(self)
        if embed:
            return embed.html

    def resolve_raw_embed(self, info, **kwargs):
        embed = get_embed_object(self)
        if embed:
            return {
                "title": embed.title,
                "type": embed.type,
                "thumbnail_url": embed.thumbnail_url,
                "width": embed.width,
                "height": embed.height,
                "html": embed.html,
            }


class StaticBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class ListBlock(graphene.ObjectType):
    items = graphene.List(graphene.NonNull(StreamFieldInterface), required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_items(self, info, **kwargs):
        # Get the nested StreamBlock type
        block_type = self.block.child_block
        # Return a list of GraphQL types from the list of values
        return [StructBlockItem(self.id, block_type, item) for item in self.value]


registry.streamfield_blocks.update(
    {
        "generic-block": StreamFieldBlock,
        blocks.CharBlock: CharBlock,
        blocks.TextBlock: TextBlock,
        blocks.EmailBlock: EmailBlock,
        blocks.IntegerBlock: IntegerBlock,
        blocks.FloatBlock: FloatBlock,
        blocks.DecimalBlock: DecimalBlock,
        blocks.RegexBlock: RegexBlock,
        blocks.URLBlock: URLBlock,
        blocks.BooleanBlock: BooleanBlock,
        blocks.DateBlock: DateBlock,
        blocks.TimeBlock: TimeBlock,
        blocks.DateTimeBlock: DateTimeBlock,
        blocks.RichTextBlock: RichTextBlock,
        blocks.RawHTMLBlock: RawHTMLBlock,
        blocks.BlockQuoteBlock: BlockQuoteBlock,
        blocks.ChoiceBlock: ChoiceBlock,
        blocks.StreamBlock: StreamBlock,
        blocks.StructBlock: StructBlock,
        blocks.StaticBlock: StaticBlock,
        blocks.ListBlock: ListBlock,
        wagtail.embeds.blocks.EmbedBlock: EmbedBlock,
    }
)


def register_streamfield_blocks():
    from .documents import get_document_type
    from .images import get_image_type
    from .pages import PageInterface
    from .snippets import SnippetTypes

    class PageChooserBlock(graphene.ObjectType):
        page = graphene.Field(PageInterface, required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_page(self, info, **kwargs):
            return self.value.specific

    class DocumentChooserBlock(graphene.ObjectType):
        document = graphene.Field(get_document_type(), required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_document(self, info, **kwargs):
            return self.value

    class ImageChooserBlock(graphene.ObjectType):
        image = graphene.Field(get_image_type(), required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_image(self, info, **kwargs):
            return self.value

    registry.streamfield_blocks.update(
        {
            blocks.PageChooserBlock: PageChooserBlock,
            wagtail.documents.blocks.DocumentChooserBlock: DocumentChooserBlock,
            wagtail.images.blocks.ImageChooserBlock: ImageChooserBlock,
        }
    )

    SnippetObjectType = SnippetTypes.get_object_type()
    if SnippetObjectType is not None:

        class SnippetChooserBlock(graphene.ObjectType):
            snippet = graphene.Field(SnippetObjectType, required=True)

            class Meta:
                interfaces = (StreamFieldInterface,)

            def resolve_snippet(self, info, **kwargs):
                return self.value

        registry.streamfield_blocks.update(
            {
                wagtail.snippets.blocks.SnippetChooserBlock: SnippetChooserBlock,
            }
        )



from typing import Optional

import graphene
from django.apps import apps

from .exceptions import IllegalDeprecation
from .registry import registry


# Classes used to define what the Django field should look like in the GQL type
class GraphQLField:
    field_name: str
    field_type: Optional[type]
    field_source: Optional[str]
    description: Optional[str]
    deprecation_reason: Optional[str]

    def __init__(
        self, field_name: str, field_type: type = None, required: bool = None, **kwargs
    ):
        # Initiate and get specific field info.
        self.field_name = field_name
        self.field_type = field_type
        self.field_source = kwargs.get("source", field_name)
        self.description = kwargs.get("description", None)
        self.deprecation_reason = kwargs.get("deprecation_reason", None)

        # Add support for NonNull/required fields
        if required:
            if self.deprecation_reason is not None:
                raise IllegalDeprecation(
                    f"Field '{field_name}' cannot be both required and deprecated"
                )
            self.field_type = graphene.NonNull(field_type, description=self.description)

        # Legacy collection API (Allow lists):
        self.extract_key = kwargs.get("key", None)
        is_list = kwargs.get("is_list", False)
        if is_list:
            self.field_type = graphene.List(
                field_type,
                required=required,
                description=self.description,
                deprecation_reason=self.deprecation_reason,
            )


def GraphQLString(field_name: str, **kwargs):
    def Mixin():
        return GraphQLField(field_name, graphene.String, **kwargs)

    return Mixin


def GraphQLFloat(field_name: str, **kwargs):
    def Mixin():
        return GraphQLField(field_name, graphene.Float, **kwargs)

    return Mixin


def GraphQLInt(field_name: str, **kwargs):
    def Mixin():
        return GraphQLField(field_name, graphene.Int, **kwargs)

    return Mixin


def GraphQLBoolean(field_name: str, **kwargs):
    def Mixin():
        return GraphQLField(field_name, graphene.Boolean, **kwargs)

    return Mixin


def GraphQLSnippet(field_name: str, snippet_model: str, **kwargs):
    def Mixin():
        from django.apps import apps

        (app_label, model) = snippet_model.lower().split(".")
        mdl = apps.get_model(app_label, model)

        if mdl:
            field_type = lambda: registry.snippets[mdl]  # noqa: E731
        else:
            field_type = graphene.String

        return GraphQLField(field_name, field_type, **kwargs)

    return Mixin


def GraphQLForeignKey(field_name, content_type, is_list=False, **kwargs):
    def Mixin():
        field_type = None
        if isinstance(content_type, str):
            app_label, model = content_type.lower().split(".")
            mdl = apps.get_model(app_label, model)
            if mdl:
                field_type = lambda: registry.models.get(mdl)  # noqa: E731
        else:
            field_type = lambda: registry.models.get(content_type)  # noqa: E731

        return GraphQLField(field_name, field_type, **kwargs)

    return Mixin


def GraphQLStreamfield(field_name: str, **kwargs):
    def Mixin():
        from .types.streamfield import StreamFieldInterface

        # Note that GraphQLStreamfield children should always be considered list elements,
        # unless they specifically are requested not to. e.g. a GraphQLStreamfield for a nested StructBlock
        if "is_list" not in kwargs:
            kwargs["is_list"] = True
        return GraphQLField(field_name, StreamFieldInterface, **kwargs)

    return Mixin


def GraphQLRichText(field_name: str, **kwargs):
    def Mixin():
        from .types.rich_text import RichText

        return GraphQLField(field_name, RichText, **kwargs)

    return Mixin


def GraphQLImage(field_name: str, **kwargs):
    def Mixin():
        from .types.images import get_image_type

        return GraphQLField(field_name, get_image_type, **kwargs)

    return Mixin


def GraphQLDocument(field_name: str, **kwargs):
    def Mixin():
        from .types.documents import get_document_type

        return GraphQLField(field_name, get_document_type, **kwargs)

    return Mixin


def GraphQLPage(field_name: str, **kwargs):
    def Mixin():
        from .types.pages import PageInterface

        return GraphQLField(field_name, PageInterface, **kwargs)

    return Mixin


def GraphQLCollection(
    nested_type,
    field_name,
    *args,
    is_queryset=False,
    is_paginated_queryset=False,
    required=False,
    description=None,
    deprecation_reason=None,
    item_required=False,
    **kwargs,
):
    def Mixin():
        from .types.structures import PaginatedQuerySet, QuerySetList

        # Check if using nested field extraction:
        source = kwargs.get("source", None)
        if source and "." in source:
            source, *key = source.split(".")
            if key:
                kwargs["source"] = source
                kwargs["key"] = key

        if required and deprecation_reason is not None:
            raise IllegalDeprecation(
                f"Field '{field_name}' cannot be both required and deprecated"
            )

        # Create the nested type
        graphql_type = nested_type(field_name, *args, required=item_required, **kwargs)

        if is_paginated_queryset:
            # Wrap the nested type in a PaginatedQuerySet type
            type_class = nested_type(field_name, *args)().field_type()
            collection_type = lambda nested_type: PaginatedQuerySet(  # noqa: E731
                nested_type,
                type_class,
                required=required,
                description=description,
                deprecation_reason=deprecation_reason,
            )
            return graphql_type, collection_type

        # Add queryset filtering when necessary.
        if (
            is_queryset
            or nested_type == GraphQLForeignKey
            or nested_type == GraphQLSnippet
        ):
            # Wrap the nested type in a QuerySetList type
            collection_type = QuerySetList
        else:
            # Wrap the nested type in a List type
            collection_type = graphene.List

        # Add support for NonNull/required to wrapper field
        required_collection_type = None
        if required:

            def required_collection_type(nested_type):
                return collection_type(nested_type, required=True)

        return graphql_type, required_collection_type or collection_type

    return Mixin


def GraphQLEmbed(field_name: str, **kwargs):
    def Mixin():
        from .types.streamfield import EmbedBlock

        return GraphQLField(field_name, EmbedBlock, **kwargs)

    return Mixin


def GraphQLTag(field_name: str, **kwargs):
    def Mixin():
        from .types.tags import TagObjectType

        if "is_list" not in kwargs:
            kwargs["is_list"] = True
        return GraphQLField(field_name, TagObjectType, **kwargs)

    return Mixin


if apps.is_installed("wagtailmedia"):

    def GraphQLMedia(field_name: str, **kwargs):
        def Mixin():
            from .types.media import get_media_type

            return GraphQLField(field_name, get_media_type, **kwargs)

        return Mixin



import graphene
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from graphene_django.types import DjangoObjectType
from graphql import GraphQLError
from wagtail.models import Page as WagtailPage
from wagtail.models import Site

from ..registry import registry
from ..utils import resolve_queryset, resolve_site_by_hostname
from .structures import QuerySetList


class PageInterface(graphene.Interface):
    id = graphene.ID()
    title = graphene.String(required=True)
    slug = graphene.String(required=True)
    content_type = graphene.String(required=True)
    page_type = graphene.String()
    live = graphene.Boolean(required=True)

    url = graphene.String()
    url_path = graphene.String(required=True)

    depth = graphene.Int()
    seo_title = graphene.String(required=True)
    search_description = graphene.String()
    show_in_menus = graphene.Boolean(required=True)

    locked = graphene.Boolean()

    first_published_at = graphene.DateTime()
    last_published_at = graphene.DateTime()

    parent = graphene.Field(lambda: PageInterface)
    children = QuerySetList(
        graphene.NonNull(lambda: PageInterface), enable_search=True, required=True
    )
    siblings = QuerySetList(
        graphene.NonNull(lambda: PageInterface), enable_search=True, required=True
    )
    next_siblings = QuerySetList(
        graphene.NonNull(lambda: PageInterface), enable_search=True, required=True
    )
    previous_siblings = QuerySetList(
        graphene.NonNull(lambda: PageInterface), enable_search=True, required=True
    )
    descendants = QuerySetList(
        graphene.NonNull(lambda: PageInterface), enable_search=True, required=True
    )
    ancestors = QuerySetList(
        graphene.NonNull(lambda: PageInterface), enable_search=True, required=True
    )

    search_score = graphene.Float()

    @classmethod
    def resolve_type(cls, instance, info, **kwargs):
        """
        If model has a custom Graphene Node type in registry then use it,
        otherwise use base page type.
        """
        return registry.pages.get(type(instance), Page)

    def resolve_content_type(self, info, **kwargs):
        self.content_type = ContentType.objects.get_for_model(self)
        return (
            self.content_type.app_label + "." + self.content_type.model_class().__name__
        )

    def resolve_page_type(self, info, **kwargs):
        return PageInterface.resolve_type(self.specific, info, **kwargs)

    def resolve_parent(self, info, **kwargs):
        """
        Resolves the parent node of current page node.
        Docs: https://docs.wagtail.io/en/stable/reference/pages/model_reference.html#wagtail.models.Page.get_parent
        """
        try:
            return self.get_parent().specific
        except GraphQLError:
            return WagtailPage.objects.none()

    def resolve_children(self, info, **kwargs):
        """
        Resolves a list of live children of this page.
        Docs: https://docs.wagtail.io/en/stable/reference/pages/queryset_reference.html#examples
        """
        return resolve_queryset(
            self.get_children().live().public().specific(), info, **kwargs
        )

    def resolve_siblings(self, info, **kwargs):
        """
        Resolves a list of sibling nodes to this page.
        Docs: https://docs.wagtail.io/en/stable/reference/pages/queryset_reference.html?highlight=get_siblings#wagtail.query.PageQuerySet.sibling_of
        """
        return resolve_queryset(
            self.get_siblings().exclude(pk=self.pk).live().public().specific(),
            info,
            **kwargs,
        )

    def resolve_next_siblings(self, info, **kwargs):
        """
        Resolves a list of direct next siblings of this page. Similar to `resolve_siblings` with sorting.
        Source: https://github.com/wagtail/wagtail/blob/master/wagtail/core/models.py#L1384
        """
        return resolve_queryset(
            self.get_next_siblings().exclude(pk=self.pk).live().public().specific(),
            info,
            **kwargs,
        )

    def resolve_previous_siblings(self, info, **kwargs):
        """
        Resolves a list of direct prev siblings of this page. Similar to `resolve_siblings` with sorting.
        Source: https://github.com/wagtail/wagtail/blob/master/wagtail/core/models.py#L1387
        """
        return resolve_queryset(
            self.get_prev_siblings().exclude(pk=self.pk).live().public().specific(),
            info,
            **kwargs,
        )

    def resolve_descendants(self, info, **kwargs):
        """
        Resolves a list of nodes pointing to the current page’s descendants.
        Docs: https://docs.wagtail.io/en/stable/reference/pages/model_reference.html#wagtail.models.Page.get_descendants
        """
        return resolve_queryset(
            self.get_descendants().live().public().specific(), info, **kwargs
        )

    def resolve_ancestors(self, info, **kwargs):
        """
        Resolves a list of nodes pointing to the current page’s ancestors.
        Docs: https://docs.wagtail.io/en/stable/reference/pages/model_reference.html#wagtail.models.Page.get_ancestors
        """
        return resolve_queryset(
            self.get_ancestors().live().public().specific(), info, **kwargs
        )

    def resolve_seo_title(self, info, **kwargs):
        """
        Get page's SEO title. Fallback to a normal page's title if absent.
        """
        return self.seo_title or self.title

    def resolve_search_score(self, info, **kwargs):
        """
        Get page's search score, will be None if not in a search context.
        """
        return getattr(self, "search_score", None)


class Page(DjangoObjectType):
    """
    Base Page type used if one isn't generated for the current model.
    All other node types extend this.
    """

    class Meta:
        model = WagtailPage
        interfaces = (PageInterface,)


def get_preview_page(token):
    """
    Get a preview page from a token.
    """
    try:
        token_params_list = token.split(":").pop(0)
        token_params_kvstr = token_params_list.split(";")

        params = {}
        for arg in token_params_kvstr:
            key, value = arg.split("=")
            params[key] = value

        if _id := params.get("id"):
            """
            This is a page that had already been saved. Lookup the class and call get_page_from_preview_token.

            TODO: update headless preview to always send page_type in the token so we can always
            the if content_type branch and eliminate the if id branch.
            """
            if page := WagtailPage.objects.get(pk=_id).specific:
                cls = type(page)
                """
                get_page_from_preview_token is added by wagtail-headless-preview,
                this condition checks that headless-preview is installed and enabled
                for the model.
                """
                if hasattr(cls, "get_page_from_preview_token"):
                    """we assume that get_page_from_preview_token validates the token"""
                    return cls.get_page_from_preview_token(token)

        if content_type := params.get("page_type"):
            """
            this is a page which has not been saved yet. lookup the content_type to get the class
            and call get_page_from_preview_token.
            """
            app_label, model = content_type.lower().split(".")
            if ctype := ContentType.objects.get(app_label=app_label, model=model):
                cls = ctype.model_class()
                """
                get_page_from_preview_token is added by wagtail-headless-preview,
                this condition checks that headless-preview is installed and enabled
                for the model.
                """
                if hasattr(cls, "get_page_from_preview_token"):
                    """we assume that get_page_from_preview_token validates the token"""
                    return cls.get_page_from_preview_token(token)
    except (WagtailPage.DoesNotExist, ContentType.DoesNotExist, ValueError):
        """
        catch and suppress errors. we don't want to expose any information about unpublished content
        accidentally.
        TODO: consider logging here.
        """
        return None


def get_specific_page(
    id=None, slug=None, url_path=None, token=None, content_type=None, site=None
):
    """
    Get a specific page, given a page_id, slug or preview if a preview token is passed
    """
    page = None
    try:
        if token:
            return get_preview_page(token)

        # Everything but the special RootPage
        qs = WagtailPage.objects.live().public().filter(depth__gt=1).specific()

        if site:
            qs = qs.in_site(site)

        if content_type:
            app_label, model = content_type.lower().split(".")
            qs = qs.filter(
                content_type=ContentType.objects.get(app_label=app_label, model=model)
            )

        if id:
            page = qs.get(pk=id)
        elif slug:
            page = qs.get(slug=slug)
        elif url_path:
            if not url_path.endswith("/"):
                url_path += "/"

            if site:
                # Got a site, so make the url_path query as specific as possible
                qs = qs.filter(
                    url_path=f"{site.root_page.url_path}{url_path.lstrip('/')}"
                )
            else:
                # if the url_path is not specific enough, or the same url_path exists under multiple
                # site roots, only the first one will be returned.
                # To-Do: make site a 1st class argument on the page query, rather than just `in_site`
                qs = qs.filter(url_path__endswith=url_path)

            if qs.exists():
                page = qs.first()

    except WagtailPage.DoesNotExist:
        page = None

    return page


def get_site_filter(info, **kwargs):
    site_hostname = kwargs.pop("site", None)
    in_current_site = kwargs.get("in_site", False)

    if site_hostname is not None and in_current_site:
        raise GraphQLError(
            "The 'site' and 'in_site' filters cannot be used at the same time."
        )

    if site_hostname is not None:
        return resolve_site_by_hostname(
            hostname=site_hostname,
            filter_name="site",
        )
    elif in_current_site:
        return Site.find_for_request(info.context)


def PagesQuery():
    # Add base type to registry
    registry.pages[type(WagtailPage)] = Page

    class Mixin:
        pages = QuerySetList(
            graphene.NonNull(lambda: PageInterface),
            content_type=graphene.Argument(
                graphene.String,
                description=_(
                    "Filter by content type. Uses the `app.Model` notation. Accepts a comma separated list of content types."
                ),
            ),
            in_site=graphene.Argument(
                graphene.Boolean,
                description=_("Filter to pages in the current site only."),
                default_value=False,
            ),
            site=graphene.Argument(
                graphene.String,
                description=_("Filter to pages in the give site."),
            ),
            ancestor=graphene.Argument(
                graphene.ID,
                description=_(
                    "Filter to pages that are descendants of the given page."
                ),
                required=False,
            ),
            parent=graphene.Argument(
                graphene.ID,
                description=_(
                    "Filter to pages that are children of the given page. "
                    "When using both `parent` and `ancestor`, then `parent` will take precendence."
                ),
                required=False,
            ),
            enable_search=True,
            required=True,
        )
        page = graphene.Field(
            PageInterface,
            id=graphene.Int(),
            slug=graphene.String(),
            url_path=graphene.Argument(
                graphene.String,
                description=_(
                    "Filter by url path. Note: in a multi-site setup, returns the first available page based. "
                    "Use `inSite: true` from the relevant site domain."
                ),
            ),
            token=graphene.Argument(
                graphene.String,
                description=_(
                    "Filter by preview token as passed by the `wagtail-headless-preview` package."
                ),
            ),
            content_type=graphene.Argument(
                graphene.String,
                description=_(
                    "Filter by content type using the app.ModelName notation. e.g. `myapp.BlogPage`"
                ),
            ),
            in_site=graphene.Argument(
                graphene.Boolean,
                description=_("Filter to pages in the current site only."),
                default_value=False,
            ),
            site=graphene.Argument(
                graphene.String,
                description=_("Filter to pages in the give site."),
            ),
        )

        # Return all pages in site, ideally specific.
        def resolve_pages(self, info, **kwargs):
            qs = WagtailPage.objects.all()

            try:
                if kwargs.get("parent"):
                    qs = WagtailPage.objects.get(id=kwargs.get("parent")).get_children()
                elif kwargs.get("ancestor"):
                    qs = WagtailPage.objects.get(
                        id=kwargs.get("ancestor")
                    ).get_descendants()
            except WagtailPage.DoesNotExist:
                qs = WagtailPage.objects.none()

            # no need to the root page
            pages = qs.live().public().filter(depth__gt=1).specific()

            site = get_site_filter(info, **kwargs)
            site_hostname = kwargs.get("site", None)
            in_current_site = kwargs.get("in_site", False)

            if site is not None:
                pages = pages.in_site(site)
            elif site is None and any(
                [site_hostname is not None, in_current_site is True]
            ):
                # If we could not resolve a Site but _were_ passed a filter, we
                # should not return any results.
                return WagtailPage.objects.none()

            content_type = kwargs.pop("content_type", None)
            content_types = content_type.split(",") if content_type else None
            if content_types:
                filters = Q()
                for content_type in content_types:
                    app_label, model = content_type.strip().lower().split(".")
                    filters |= Q(
                        content_type__app_label=app_label, content_type__model=model
                    )
                pages = pages.filter(filters)

            return resolve_queryset(pages, info, **kwargs)

        # Return a specific page, identified by ID or Slug.
        def resolve_page(self, info, **kwargs):
            return get_specific_page(
                id=kwargs.get("id"),
                slug=kwargs.get("slug"),
                url_path=kwargs.get("url_path"),
                token=kwargs.get("token"),
                content_type=kwargs.get("content_type"),
                site=get_site_filter(info, **kwargs),
            )

    return Mixin



from typing import Literal, Optional

from django.conf import settings
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.db import connection
from graphql import GraphQLError
from wagtail.models import Site
from wagtail.search.index import class_is_indexed
from wagtail.search.models import Query

from .settings import grapple_settings
from .types.structures import BasePaginatedType, PaginationType


def resolve_site_by_id(
    *,
    id: int,
) -> Optional[Site]:
    """
    Find a `Site` object by ID
    """

    try:
        return Site.objects.get(id=id)
    except Site.DoesNotExist:
        # This is an expected error, so should not raise a GraphQLError.
        return None


def resolve_site_by_hostname(
    *,
    hostname: str,
    filter_name: Literal["site", "hostname"],
) -> Optional[Site]:
    """
    Find a `Site` object by hostname.

    If two `Site` records exist with the same hostname, you must include the
    port to disambiguate.

    For example:

    >>> resolve_site_by_hostname(hostname="wagtail.org")
    >>> resolve_site_by_hostname(hostname="wagtail.org:443")
    """

    # Optionally allow querying by port
    if ":" in hostname:
        (hostname, port) = hostname.split(":", 1)
        query = {
            "hostname": hostname,
            "port": port,
        }
    else:
        query = {
            "hostname": hostname,
        }

    try:
        return Site.objects.get(**query)
    except Site.MultipleObjectsReturned as err:
        raise GraphQLError(
            f"Your filter `{filter_name}={hostname}` returned "
            "multiple sites. Try including a port number to disambiguate "
            f"(e.g. `{filter_name}={hostname}:8000`)."
        ) from err
    except Site.DoesNotExist:
        # This is an expected error, so should not raise a GraphQLError.
        return None


def _sliced_queryset(qs, limit=None, offset=None):
    offset = int(offset or 0)
    # default
    limit = int(limit or grapple_settings.PAGE_SIZE)
    # maximum
    limit = min(limit, grapple_settings.MAX_PAGE_SIZE)
    return qs[offset : limit + offset]


def resolve_queryset(
    qs,
    info,
    limit=None,
    offset=None,
    search_query=None,
    id=None,
    order=None,
    collection=None,
    **kwargs,
):
    """
    Add limit, offset and search capabilities to the query. This contains
    argument names used by
    :class:`~grapple.types.structures.QuerySetList`.
    :param qs: The query set to be modified.
    :param info: The Graphene info object.
    :param limit: Limit number of objects in the QuerySet.
    :type limit: int
    :param id: Filter by the primary key.
    :type id: int
    :param offset: Omit a number of objects from the beginning of the query set
    :type offset: int
    :param search_query: Using Wagtail search, exclude objects that do not match
                         the search query.
    :type search_query: str
    :param order: Order the query set using the Django QuerySet order_by format.
    :type order: str
    :param collection: Use Wagtail's collection id to filter images or documents
    :type collection: int
    """

    if id is not None:
        qs = qs.filter(pk=id)
    else:
        qs = qs.all()

    order_by_relevance = True
    if order is not None:
        qs = qs.order_by(*(x.strip() for x in order.split(",")))
        order_by_relevance = False

    if collection is not None:
        try:
            qs.model._meta.get_field("collection")
        except LookupError:
            pass
        else:
            qs = qs.filter(collection=collection)

    if id is None and search_query:
        # Check if the queryset is searchable using Wagtail search.
        if not class_is_indexed(qs.model):
            raise TypeError("This data type is not searchable by Wagtail.")

        if grapple_settings.ADD_SEARCH_HIT:
            query = Query.get(search_query)
            query.add_hit()

        qs = qs.search(search_query, order_by_relevance=order_by_relevance)
        if connection.vendor != "sqlite":
            qs = qs.annotate_score("search_score")

    return _sliced_queryset(qs, limit, offset)


def get_paginated_result(qs, page, per_page):
    """
    Returns a paginated result.
    """
    paginator = Paginator(qs, per_page)

    try:
        # If the page exists and the page is an int
        page_obj = paginator.page(page)
    except PageNotAnInteger:
        # If the page is not an int; show the first page
        page_obj = paginator.page(1)
    except EmptyPage:
        # If the page is out of range (too high most likely)
        # Then return the last page
        page_obj = paginator.page(paginator.num_pages)

    return BasePaginatedType(
        items=page_obj.object_list,
        pagination=PaginationType(
            total=paginator.count,
            count=len(page_obj.object_list),
            per_page=per_page,
            current_page=page_obj.number,
            prev_page=page_obj.previous_page_number()
            if page_obj.has_previous()
            else None,
            next_page=page_obj.next_page_number() if page_obj.has_next() else None,
            total_pages=paginator.num_pages,
        ),
    )


def resolve_paginated_queryset(
    qs, info, page=None, per_page=None, search_query=None, id=None, order=None, **kwargs
):
    """
    Add page, per_page and search capabilities to the query. This contains
    argument names used by
    :function:`~grapple.types.structures.PaginatedQuerySet`.
    :param qs: The query set to be modified.
    :param info: The Graphene info object.
    :param page: Page of resulting objects to return from the QuerySet.
    :type page: int
    :param id: Filter by the primary key.
    :type id: int
    :param per_page: The maximum number of items to include on a page.
    :type per_page: int
    :param search_query: Using Wagtail search, exclude objects that do not match
                         the search query.
    :type search_query: str
    :param order: Order the query set using the Django QuerySet order_by format.
    :type order: str
    """
    page = int(page or 1)
    per_page = min(
        int(per_page or grapple_settings.PAGE_SIZE), grapple_settings.MAX_PAGE_SIZE
    )

    if id is not None:
        qs = qs.filter(pk=id)
    else:
        qs = qs.all()

    # order_by_relevance will always take precedence over an existing order_by in the Postgres backend
    # we need to set it to False if we want to specify our own order_by.
    order_by_relevance = True
    if order is not None:
        qs = qs.order_by(*(x.strip() for x in order.split(",")))
        order_by_relevance = False

    if id is None and search_query:
        # Check if the queryset is searchable using Wagtail search.
        if not class_is_indexed(qs.model):
            raise TypeError("This data type is not searchable by Wagtail.")

        if grapple_settings.ADD_SEARCH_HIT:
            query = Query.get(search_query)
            query.add_hit()

        qs = qs.search(search_query, order_by_relevance=order_by_relevance)
        if connection.vendor != "sqlite":
            qs = qs.annotate_score("search_score")

    return get_paginated_result(qs, page, per_page)


def get_media_item_url(cls):
    url = ""
    if hasattr(cls, "url"):
        url = cls.url
    elif hasattr(cls, "file"):
        url = cls.file.url

    if url[0] == "/":
        return settings.BASE_URL + url
    return url



import inspect
from collections.abc import Iterable
from types import MethodType
from typing import Any, Dict, Type, Union

import graphene
from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from graphene_django.types import DjangoObjectType
from wagtail.blocks import StructValue, stream_block
from wagtail.contrib.settings.models import BaseGenericSetting, BaseSiteSetting
from wagtail.documents.models import AbstractDocument
from wagtail.fields import RichTextField
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import AbstractImage, AbstractRendition
from wagtail.models import Page as WagtailPage
from wagtail.rich_text import RichText
from wagtail.snippets.models import get_snippet_models

from .helpers import field_middlewares, streamfield_types
from .registry import registry
from .settings import grapple_settings
from .types.documents import DocumentObjectType
from .types.images import ImageObjectType, ImageRenditionObjectType
from .types.pages import Page, PageInterface
from .types.rich_text import RichText as RichTextType
from .types.streamfield import generate_streamfield_union

if apps.is_installed("wagtailmedia"):
    from wagtailmedia.models import AbstractMedia

    from .types.media import MediaObjectType

    has_wagtail_media = True

else:
    # TODO: find a better way to have this as an optional dependency
    class AbstractMedia:
        def __init__(self):
            pass

        def __add__(self, other):
            pass

        def __name__(self):
            pass

    MediaObjectType = None
    has_wagtail_media = False


def import_apps():
    """
    Add each django app set in the settings file
    """

    # Register each app in the django project.
    if isinstance(grapple_settings.APPS, (list, tuple)):
        for name in grapple_settings.APPS:
            add_app(name)
            registry.apps.append(name)
    else:
        apps = grapple_settings.APPS.items()
        for name, prefix in apps:
            add_app(name, prefix)
            registry.apps.append(name)

    # Register any 'decorated' StreamField structs.
    for streamfield_type in streamfield_types:
        cls = streamfield_type["cls"]
        base_type = streamfield_type["base_type"]

        if hasattr(cls, "graphql_types"):
            base_type = generate_streamfield_union(cls.graphql_types)

        node_type = build_streamfield_type(
            cls,
            streamfield_type["type_prefix"],
            streamfield_type["interface"],
            base_type,
        )

        registry.streamfield_blocks[streamfield_type["cls"]] = node_type

    registry.field_middlewares = field_middlewares


def add_app(app_label: str, prefix: str = ""):
    """
    Iterate through each model in the app and pass it to node type creators.
    """
    from django.apps import apps

    # Get the required django app.
    app = apps.get_app_config(app_label)

    # Create a collection of models of standard models (Pages, Images, Documents).
    models = list(app.get_models())

    # Add snippet models to the model collection.
    for snippet in get_snippet_models():
        if snippet._meta.app_label == app_label:
            models.append(snippet)

    # Create add each model to correct section of registry.
    for model in models:
        register_model(model, prefix)


def register_model(cls: type, type_prefix: str):
    """
    Pass model to the right node type creator based on it's base class.
    """

    # Pass class to correct type creator.
    if cls is not None:
        if issubclass(cls, WagtailPage):
            register_page_model(cls, type_prefix)
        elif issubclass(cls, AbstractDocument):
            register_document_model(cls, type_prefix)
        elif issubclass(cls, AbstractImage):
            register_image_model(cls, type_prefix)
        elif issubclass(cls, AbstractRendition):
            register_image_rendition_model(cls, type_prefix)
        elif has_wagtail_media and issubclass(cls, AbstractMedia):
            register_media_model(cls, type_prefix)
        elif issubclass(cls, (BaseSiteSetting, BaseGenericSetting)):
            register_settings_model(cls, type_prefix)
        elif cls in get_snippet_models():
            register_snippet_model(cls, type_prefix)
        else:
            register_django_model(cls, type_prefix)


def get_fields_and_properties(cls):
    """
    Return all fields and @property methods for a model.
    """
    from graphene_django.utils import get_model_fields

    # Note: graphene-django use this method to get the model fields
    # cls._meta.get_fields(include_parents=False) includes symmetrical ManyToMany fields, while get_model_fields doesn't
    fields = [field for field, instance in get_model_fields(cls)]

    properties = [
        method[0]
        for method in inspect.getmembers(cls, lambda o: isinstance(o, property))
    ]

    return fields + properties


def get_field_type(field):
    # If a tuple is returned then obj[1] wraps obj[0]
    field_wrapper = None
    if isinstance(field, tuple):
        field, field_wrapper = field
        if callable(field):
            field = field()

    field_type = field.field_type
    if field_type is not None:
        if field_wrapper:
            return field, field_wrapper(field_type)
        else:
            return field, graphene.Field(
                field_type,
                description=field.description,
                deprecation_reason=field.deprecation_reason,
            )


def model_resolver(field):
    def mixin(self, instance, info, **kwargs):
        from .utils import resolve_queryset

        cls_field = getattr(instance, field.field_source)

        # If queryset then call .all() method
        if issubclass(type(cls_field), models.Manager):
            # Shortcut to extract one nested field from an list of objects
            def get_nested_field(cls, extract_key):
                # If last value in list then return that from the class.
                if len(extract_key) == 1:
                    return getattr(cls, extract_key[0])

                # Get data from nested field
                field = getattr(cls, extract_key[0])
                if field is None:
                    return None
                if issubclass(type(field), models.Manager):
                    field = field.all()

                # If field data is a list then iterate over it
                if isinstance(field, Iterable):
                    return [
                        get_nested_field(nested_cls, extract_key[1:])
                        for nested_cls in field
                    ]

                # If single value then return it.
                return get_nested_field(field, extract_key[1:])

            if field.extract_key:
                return [
                    get_nested_field(cls, field.extract_key) for cls in cls_field.all()
                ]

            # Check if any queryset params:
            if not kwargs:
                return cls_field.all()
            return resolve_queryset(cls_field, info, **kwargs)

        # If method then call and return result
        if callable(cls_field):
            return cls_field(info, **kwargs)

        # Expand HTML if the value's field is richtext
        if field.field_type is RichTextType:
            # Rendering of html will be handled by the GraphQL executor calling
            # RichText.serialize, due to being declared as GraphQLRichText rather than
            # GraphQLString
            return cls_field
        try:
            if hasattr(instance._meta, "get_field"):
                field_model = instance._meta.get_field(field.field_source)
            else:
                field_model = instance._meta.fields[field.field_source]
        except FieldDoesNotExist:
            return cls_field

        if type(field_model) is RichTextField:
            return RichTextType.serialize(cls_field)

        # If none of those then just return field
        return cls_field

    return mixin


def build_node_type(
    cls: type,
    type_prefix: str,
    interface: graphene.Interface,
    base_type: Type[DjangoObjectType] = DjangoObjectType,
):
    """
    Build a graphene node type from a model class and associate
    with an interface. If it has custom fields then implement them.
    """
    type_name = type_prefix + cls.__name__

    # Create a temporary model and temporary node that will be replaced later on.
    class UnmanagedMeta:
        app_label = type_name
        managed = False

    stub_model = type(
        type_name, (models.Model,), {"__module__": "", "Meta": UnmanagedMeta}
    )

    class StubMeta:
        model = stub_model

    type_meta = {
        "Meta": StubMeta,
        "type": lambda: {
            "cls": cls,
            "lazy": True,
            "name": type_name,
            "base_type": base_type,
            "interface": interface,
        },
    }

    return type("Stub" + type_name, (DjangoObjectType,), type_meta)


def load_type_fields():
    for list_name in registry.lazy_types:
        type_list = getattr(registry, list_name)

        for key, node in type_list.items():
            type_definition = node.type()
            if type_definition.get("lazy"):
                # Get the original django model data
                cls = type_definition.get("cls")
                base_type = type_definition.get("base_type")
                interface = type_definition.get("interface")
                type_name = type_definition.get("name")

                # Recreate the graphene type with the fields set
                class Meta:
                    model = cls
                    interfaces = (interface,) if interface is not None else ()

                type_meta = {"Meta": Meta, "id": graphene.ID(), "name": type_name}

                exclude_fields = []
                base_type_for_exclusion_checks = (
                    base_type if not issubclass(cls, WagtailPage) else WagtailPage
                )
                for field in get_fields_and_properties(cls):
                    # Filter out any fields that are defined on the interface of base type to prevent the
                    # 'Excluding the custom field "<field>" on DjangoObjectType "<cls>" has no effect.
                    # Either remove the custom field or remove the field from the "exclude" list.' warning
                    if (
                        field == "id"
                        or hasattr(interface, field)
                        or hasattr(base_type_for_exclusion_checks, field)
                    ):
                        continue

                    exclude_fields.append(field)

                # Add any custom fields to node if they are defined.
                methods = {}
                if hasattr(cls, "graphql_fields"):
                    for field in cls.graphql_fields:
                        if callable(field):
                            field = field()

                        # Add field to GQL type with correct field-type
                        field, field_type = get_field_type(field)
                        type_meta[field.field_name] = field_type

                        # Remove field from excluded list
                        if field.field_name in exclude_fields:
                            exclude_fields.remove(field.field_name)

                        # Add a custom resolver for each field
                        methods["resolve_" + field.field_name] = model_resolver(field)

                # Replace stud node with real thing
                type_meta["Meta"].exclude_fields = exclude_fields
                node = type(type_name, (base_type,), type_meta)

                # Add custom resolvers for fields
                for name, method in methods.items():
                    setattr(node, name, MethodType(method, node))

                # Update list with new node
                type_list[key] = node


def convert_to_underscore(name):
    import re

    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def get_field_value(instance, field_name: str):
    """
    Returns the value of a given field on an object of a StreamField.

    Different types of objects require different ways to access the values.
    """
    if isinstance(instance, StructValue):
        return instance[field_name]
    elif isinstance(instance.value, RichText):
        return RichTextType.serialize(instance.value.source)
    elif isinstance(instance.value, stream_block.StreamValue):
        stream_data = dict(instance.value.stream_data)
        return stream_data[field_name]
    else:
        return instance.value[field_name]


def get_all_field_values(*, instance, cls) -> Dict[str, Any]:
    """
    Returns a dictionary of all fields and their values within a given stream
    block instance.
    """

    values: Dict[str, Any] = {}
    for item in cls.base_blocks.items():
        field_name: str = item[0]
        values[field_name] = get_field_value(
            instance=instance,
            field_name=field_name,
        )

    return values


def streamfield_resolver(self, instance, info, **kwargs):
    value = None
    if hasattr(instance, "block"):
        field_name = convert_to_underscore(info.field_name)
        block = instance.block.child_blocks[field_name]

        if not block:
            return None

        value = get_field_value(instance, field_name)
        if issubclass(type(block), ImageChooserBlock) and isinstance(value, int):
            return block.to_python(value)

    return value


def custom_cls_resolver(*, cls, graphql_field):
    klass = cls()

    # If we've defined a `source` kwarg: use it.
    if hasattr(graphql_field, "field_source") and hasattr(
        klass, graphql_field.field_source
    ):
        if isinstance(getattr(type(cls()), graphql_field.field_source), property):
            return lambda self, instance, info, **kwargs: getattr(
                klass, graphql_field.field_source
            )
        elif callable(getattr(cls, graphql_field.field_source)):
            return lambda self, instance, info, **kwargs: getattr(
                klass, graphql_field.field_source
            )(values=get_all_field_values(instance=instance, cls=cls), **kwargs)

    # If the `field_name` is a property or method of the class: use it.
    if hasattr(graphql_field, "field_name") and hasattr(
        klass, graphql_field.field_name
    ):
        if isinstance(getattr(type(cls()), graphql_field.field_name), property):
            return lambda self, instance, info, **kwargs: getattr(
                klass, graphql_field.field_name
            )
        elif callable(getattr(cls, graphql_field.field_name)):
            return lambda self, instance, info, **kwargs: getattr(
                klass, graphql_field.field_name
            )(values=get_all_field_values(instance=instance, cls=cls), **kwargs)

    # No match found - fall back to the streamfield_resolver() later.
    return None


def build_streamfield_type(
    cls: type,
    type_prefix: str,
    interface: graphene.Interface,
    base_type=graphene.ObjectType,
):
    """
    Build a graphql type for a StreamBlock or StructBlock class
    If it has custom fields then implement them.
    """

    # Create a new blank node type
    class Meta:
        if hasattr(cls, "graphql_types"):
            types = [
                registry.streamfield_blocks.get(block) for block in cls.graphql_types
            ]
        else:
            interfaces = (interface,) if interface is not None else ()
        # Add description to type if the Meta class declares it
        description = getattr(cls._meta_class, "graphql_description", None)

    methods = {}
    type_name = type_prefix + cls.__name__
    type_meta = {"Meta": Meta, "id": graphene.String()}

    # Add any custom fields to node if they are defined.
    if hasattr(cls, "graphql_fields"):
        for item in cls.graphql_fields:
            if callable(item):
                item = item()

            # Get correct types from field
            field, field_type = get_field_type(item)

            # Add support for `graphql_fields`
            methods["resolve_" + field.field_name] = (
                custom_cls_resolver(cls=cls, graphql_field=field)
                or streamfield_resolver
            )

            # Add field to GQL type with correct field-type
            type_meta[field.field_name] = field_type

    # Set excluded fields to stop errors cropping up from unsupported field types.
    graphql_node = type(type_name, (base_type,), type_meta)

    for name, method in methods.items():
        setattr(graphql_node, name, MethodType(method, graphql_node))

    return graphql_node


def register_page_model(cls: Type[WagtailPage], type_prefix: str):
    """
    Create graphene node type for models than inherit from Wagtail Page model.
    """

    # Avoid gql type duplicates
    if cls in registry.pages:
        return

    # Create a GQL type derived from the page model.
    if page_node_type := build_node_type(cls, type_prefix, PageInterface, Page):
        # Add page type to registry.
        registry.pages[cls] = page_node_type


def register_document_model(cls: Type[AbstractDocument], type_prefix: str):
    """
    Create graphene node type for a model than inherits from AbstractDocument.
    Only one model will actually be generated because a default document model
    needs to be set in settings.
    """

    # Avoid gql type duplicates
    if cls in registry.documents:
        return

    # Create a GQL type derived from the document model.
    if document_node_type := build_node_type(
        cls, type_prefix, None, DocumentObjectType
    ):
        # Add document type to registry.
        registry.documents[cls] = document_node_type


def register_image_model(cls: Type[AbstractImage], type_prefix: str):
    """
    Create a graphene node type for a model than inherits from AbstractImage.
    Only one type will actually be generated because a default image model
    needs to be set in settings.
    """

    # Avoid gql type duplicates
    if cls in registry.images:
        return

    # Create a GQL type derived from the image model.
    if image_node_type := build_node_type(cls, type_prefix, None, ImageObjectType):
        # Add image type to registry.
        registry.images[cls] = image_node_type


def register_image_rendition_model(cls: Type[AbstractRendition], type_prefix: str):
    """
    Create a graphene node type for a model than inherits from AbstractImageRendition.
    Only one type will actually be generated because a default image model
    needs to be set in settings.
    """

    if cls in registry.images:
        return

    # Create a GQL type derived from the image rendition model.
    if rendition_type := build_node_type(
        cls, type_prefix, None, ImageRenditionObjectType
    ):
        registry.images[cls] = rendition_type


def register_media_model(cls: Type[AbstractMedia], type_prefix: str):
    """
    Create graphene node type for a model than inherits from AbstractDocument.
    Only one model will actually be generated because a default document model
    needs to be set in settings.
    """

    if cls in registry.media:
        return

    # Create a GQL type derived from the media model.
    if media_node_type := build_node_type(cls, type_prefix, None, MediaObjectType):
        registry.media[cls] = media_node_type


def register_settings_model(
    cls: Union[Type[BaseSiteSetting], Type[BaseGenericSetting]], type_prefix: str
):
    """
    Create a graphene node type for a settings page.
    """

    if cls in registry.settings:
        return

    # Create a GQL type that for the Settings model
    if settings_node_type := build_node_type(cls, type_prefix, None):
        registry.settings[cls] = settings_node_type


def register_snippet_model(cls: Type[models.Model], type_prefix: str):
    """
    Create a graphene type for a snippet model.
    """

    if cls in registry.snippets:
        return

    # Create a GQL type that implements Snippet Interface
    snippet_node_type = build_node_type(cls, type_prefix, None)

    if snippet_node_type:
        registry.snippets[cls] = snippet_node_type


def register_django_model(cls: Type[models.Model], type_prefix: str):
    """
    Create a graphene type for (non-specific) Django model.
    Used for Orderables and other foreign keys.
    """

    # Avoid gql type duplicates
    if cls in registry.django_models:
        return

    # Create a GQL type for the non-specific Django model.
    if django_node_type := build_node_type(cls, type_prefix, None):
        registry.django_models[cls] = django_node_type



import graphene
from graphene_django import DjangoObjectType
from wagtail.images import get_image_model
from wagtail.images.models import Image as WagtailImage
from wagtail.images.models import Rendition as WagtailImageRendition
from wagtail.images.models import SourceImageIOError

from ..registry import registry
from ..settings import grapple_settings
from ..utils import get_media_item_url, resolve_queryset
from .collections import CollectionObjectType
from .structures import QuerySetList
from .tags import TagObjectType


def get_image_type():
    return registry.images.get(get_image_model(), ImageObjectType)


def get_rendition_type():
    rendition_mdl = get_image_model().renditions.rel.related_model
    return registry.images.get(rendition_mdl, ImageRenditionObjectType)


def rendition_allowed(rendition_filter):
    """Checks a given rendition filter is allowed"""
    allowed_filters = grapple_settings.ALLOWED_IMAGE_FILTERS
    if allowed_filters is None or not isinstance(allowed_filters, (list, tuple)):
        return True

    return rendition_filter in allowed_filters


class ImageRenditionObjectType(DjangoObjectType):
    id = graphene.ID(required=True)
    file = graphene.String(required=True)
    image = graphene.Field(lambda: get_image_type(), required=True)
    filter_spec = graphene.String(required=True)
    width = graphene.Int(required=True)
    height = graphene.Int(required=True)
    focal_point_key = graphene.String(required=True)
    focal_point = graphene.String()
    url = graphene.String(required=True)
    alt = graphene.String(required=True)
    background_position_style = graphene.String(required=True)

    class Meta:
        model = WagtailImageRendition

    def resolve_url(instance, info, **kwargs):
        return instance.full_url


class ImageObjectType(DjangoObjectType):
    id = graphene.ID(required=True)
    title = graphene.String(required=True)
    file = graphene.String(required=True)
    width = graphene.Int(required=True)
    height = graphene.Int(required=True)
    created_at = graphene.DateTime(required=True)
    focal_point_x = graphene.Int()
    focal_point_y = graphene.Int()
    focal_point_width = graphene.Int()
    focal_point_height = graphene.Int()
    file_size = graphene.Int()
    file_hash = graphene.String(required=True)
    src = graphene.String(required=True, deprecation_reason="Use the `url` attribute")
    url = graphene.String(required=True)
    aspect_ratio = graphene.Float(required=True)
    sizes = graphene.String(required=True)
    collection = graphene.Field(lambda: CollectionObjectType, required=True)
    tags = graphene.List(graphene.NonNull(lambda: TagObjectType), required=True)
    rendition = graphene.Field(
        lambda: get_rendition_type(),
        max=graphene.String(),
        min=graphene.String(),
        width=graphene.Int(),
        height=graphene.Int(),
        fill=graphene.String(),
        format=graphene.String(),
        bgcolor=graphene.String(),
        jpegquality=graphene.Int(),
        webpquality=graphene.Int(),
    )
    src_set = graphene.String(
        sizes=graphene.List(graphene.Int), format=graphene.String()
    )

    class Meta:
        model = WagtailImage

    def resolve_rendition(instance, info, **kwargs):
        """
        Render a custom rendition of the current image.
        """
        filters = "|".join([f"{key}-{val}" for key, val in kwargs.items()])

        # Only allowed the defined filters (thus renditions)
        if not rendition_allowed(filters):
            return
        try:
            return instance.get_rendition(filters)
        except SourceImageIOError:
            return

    def resolve_url(instance, info, **kwargs):
        """
        Get the uploaded image url.
        """
        return get_media_item_url(instance)

    def resolve_src(self, info, **kwargs):
        """
        Deprecated. Use the `url` attribute.
        """
        return get_media_item_url(self)

    def resolve_aspect_ratio(instance, info, **kwargs):
        """
        Calculate aspect ratio for the image.
        """
        return instance.width / instance.height

    def resolve_sizes(instance, info, **kwargs):
        return f"(max-width: {instance.width}px) 100vw, {instance.width}px"

    def resolve_tags(instance, info, **kwargs):
        return instance.tags.all()

    def resolve_src_set(instance, info, sizes, format=None, **kwargs):
        """
        Generate src set of renditions.
        """
        filter_suffix = f"|format-{format}" if format else ""
        format_kwarg = {"format": format} if format else {}
        if instance.file.name is not None:
            rendition_list = [
                ImageObjectType.resolve_rendition(
                    instance, info, width=width, **format_kwarg
                )
                for width in sizes
                if rendition_allowed(f"width-{width}{filter_suffix}")
            ]

            return ", ".join(
                [
                    f"{get_media_item_url(img)} {img.width}w"
                    for img in rendition_list
                    if img is not None
                ]
            )

        return ""


def ImagesQuery():
    mdl = get_image_model()
    mdl_type = get_image_type()

    class Mixin:
        image = graphene.Field(mdl_type, id=graphene.ID())
        images = QuerySetList(
            graphene.NonNull(mdl_type),
            enable_search=True,
            required=True,
            collection=graphene.Argument(
                graphene.ID, description="Filter by collection id"
            ),
        )
        image_type = graphene.String(required=True)

        def resolve_image(parent, info, id, **kwargs):
            """Returns an image given the id, if in a public collection"""
            try:
                return (
                    mdl.objects.filter(collection__view_restrictions__isnull=True)
                    .prefetch_renditions()
                    .get(pk=id)
                )
            except mdl.DoesNotExist:
                return None

        def resolve_images(parent, info, **kwargs):
            """Returns all images in a public collection"""
            return resolve_queryset(
                mdl.objects.filter(
                    collection__view_restrictions__isnull=True
                ).prefetch_renditions(),
                info,
                **kwargs,
            )

        # Give name of the image type, used to generate mixins
        def resolve_image_type(parent, info, **kwargs):
            return mdl_type

    return Mixin



from functools import partial

from graphene import ResolveInfo
from graphql.execution.middleware import get_middleware_resolvers

from .registry import registry

ROOT_TYPES = ["Query", "Mutation", "Subscription"]


class IsAuthenticatedMiddleware:
    def resolve(self, next, root, info, **kwargs):
        if not info.context.user.is_authenticated:
            return None
        return next(root, info, **kwargs)


class IsAnonymousMiddleware:
    def resolve(self, next, root, info, **kwargs):
        if not info.context.user.is_anonymous:
            return None
        return next(root, info, **kwargs)


class GrappleMiddleware:
    def __init__(self):
        self.field_middlewares = {}
        for field_name in registry.field_middlewares:
            #  get_middleware_resolvers expects an instantiated class, if passing a class-based middleware.
            middlewares = tuple(
                middleware() if isinstance(middleware, type) else middleware
                for middleware in registry.field_middlewares[field_name]
            )
            self.field_middlewares[field_name] = list(
                get_middleware_resolvers(middlewares)
            )

    def resolve(self, next, root, info: ResolveInfo, **kwargs):
        field_name = info.field_name
        parent_name = info.parent_type.name
        if field_name in self.field_middlewares and parent_name in ROOT_TYPES:
            middlewares = self.field_middlewares[field_name].copy()
            while middlewares:
                next = partial(middlewares.pop(), next)

        return next(root, info, **kwargs)



"""
Settings for Wagtail Grapple are all namespaced in the GRAPPLE setting.
For example your project's `settings.py` file might look like this:
GRAPPLE = {
    'APPS': ['home'],
    'ADD_SEARCH_HIT': True,
}
This module provides the `grapple_settings` object, that is used to access
Wagtail Grapple settings, checking for user settings first, then falling
back to the defaults.
"""
import logging

from django.conf import settings as django_settings
from django.test.signals import setting_changed

logger = logging.getLogger("grapple")


DEFAULTS = {
    "APPS": [],
    "AUTO_CAMELCASE": True,
    "ALLOWED_IMAGE_FILTERS": None,
    "EXPOSE_GRAPHIQL": False,
    "ADD_SEARCH_HIT": False,
    "PAGE_SIZE": 10,
    "MAX_PAGE_SIZE": 100,
    "RICHTEXT_FORMAT": "html",
}

# List of settings that have been deprecated
DEPRECATED_SETTINGS = [
    "GRAPPLE_APPS",
    "GRAPPLE_ADD_SEARCH_HIT",
    "GRAPPLE_AUTO_CAMELCASE",
    "GRAPPLE_EXPOSE_GRAPHIQL",
    "GRAPPLE_ALLOWED_IMAGE_FILTERS",
]

# List of settings that have been removed
REMOVED_SETTINGS = []


class GrappleSettings:
    """
    A settings object that allows Wagtail Grapple settings to be accessed as
    properties. For example:
        from grapple.settings import grapple_settings
        print(grapple_settings.APPS)
    Note:
    This is an internal class that is only compatible with settings namespaced
    under the GRAPPLE name. It is not intended to be used by 3rd-party
    apps, and test helpers like `override_settings` may not work as expected.
    """

    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)
        self.defaults = defaults or DEFAULTS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = self.__check_user_settings(
                getattr(django_settings, "GRAPPLE", {})
            )
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Grapple setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def __check_user_settings(self, user_settings):
        for setting in DEPRECATED_SETTINGS:
            if setting in user_settings or hasattr(django_settings, setting):
                new_setting = setting.replace("GRAPPLE_", "")
                logger.warning(
                    f"The '{setting}' setting is deprecated and will be removed in the next release, use GRAPPLE['{new_setting}'] instead."
                )
                if setting in user_settings:
                    user_settings[new_setting] = user_settings[setting]
                else:
                    user_settings[new_setting] = getattr(django_settings, setting)

        settings_doc_url = "https://wagtail-grapple.readthedocs.io/en/latest/general-usage/settings.html"
        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError(
                    f"The '{setting}' setting has been removed. Please refer to '{settings_doc_url}' for available settings."
                )
        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


grapple_settings = GrappleSettings(None, DEFAULTS)


def reload_grapple_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "GRAPPLE":
        grapple_settings.reload()


setting_changed.connect(reload_grapple_settings)



# Generated by Django 2.2.1 on 2019-06-13 23:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [("contenttypes", "0002_remove_content_type_name")]

    operations = [
        migrations.CreateModel(
            name="PagePreview",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("token", models.CharField(max_length=255, unique=True)),
                ("content_json", models.TextField()),
                ("created_at", models.DateField(auto_now_add=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
        )
    ]
